﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using System.Reflection;

namespace ValidateIndex
{
    public static class Interaction
    {
        public static PropertyInfo GetPropertyInfo<T>(Expression<Func<T>> propertyExpression)
        {
            var memberExpression = propertyExpression.Body as MemberExpression;
            var property = memberExpression.Member as PropertyInfo;
            return property;
        }
        public static object GetPropertyValue<T>(object obj, Expression<Func<T>> propertyExpression)
        {
            var memberExpression = propertyExpression.Body as MemberExpression;
            var property = memberExpression.Member as PropertyInfo;
            var getMethod = property.GetGetMethod(true);
            return getMethod.Invoke(obj, null);
        }
        public static void SetPropertyValue<T>(object obj, Expression<Func<T>> propertyExpression, object value)
        {
            var memberExpression = propertyExpression.Body as MemberExpression;
            var property = memberExpression.Member as PropertyInfo;
            var setMethod = property.GetSetMethod(true);
            setMethod.Invoke(obj, new[] { value });
        }

        public static string GetPropertyName<T>(Expression<Func<T>> exp)
        {
            return (((MemberExpression)(exp.Body)).Member).Name;
        }

        // requires explicit specification of both object type and property type
        public static string GetPropertyName<TObject, TResult>(Expression<Func<TObject, TResult>> exp)
        {
            // extract property name
            return (((MemberExpression)(exp.Body)).Member).Name;
        }

        // requires explicit specification of object type
        public static string GetPropertyName<TObject>(Expression<Func<TObject, object>> exp)
        {
            var e = (exp.Body is UnaryExpression) ? (((UnaryExpression)(exp.Body)).Operand) : exp.Body;
            return (((MemberExpression)(e)).Member).Name;
        }

        public static Type GetEnumerableArgumentType(object collectionAsObject)
        {
            return (from iType in collectionAsObject.GetType().GetInterfaces() where iType.IsGenericType && iType.GetGenericTypeDefinition() == typeof(IEnumerable<>) select iType.GetGenericArguments()[0]).FirstOrDefault();
        }

        public static void CopyParams<T>(T source, T destination)
        {
            var t = typeof(T);
            var destType = destination.GetType();
            foreach (var propSource in source.GetType().GetProperties(BindingFlags.Instance | BindingFlags.Public))
            {
                if (t.GetProperty(propSource.Name) == null) continue;
                var propDest = destType.GetProperty(propSource.Name);
                if (propDest == null) continue;
                var getter = propDest.GetGetMethod(true);
                var destVal = getter.Invoke(destination, new object[] { });
                if (destVal is string && !string.IsNullOrWhiteSpace(destVal as string)) continue;

                var setter = propDest.GetSetMethod(true);
                getter = propSource.GetGetMethod(true);
                if (setter != null && getter != null)
                {
                    var value = getter.Invoke(source, new object[] { });
                    setter.Invoke(destination, new[] { value });
                }
            }
        }

        public static IEnumerable<string> GetChangedProperties<T>(T source, T destination)
        {
            var returnProps = new List<string>();
            var props = typeof(T).GetProperties();
            foreach (var prop in props)
            {
                var sourceValue = prop.GetValue(source, null);
                var destValue = prop.GetValue(destination, null);
                if (sourceValue == null && destValue == null)
                    continue;
                if (sourceValue == null)
                {
                    returnProps.Add(prop.Name);
                    continue;
                }

                if (!sourceValue.Equals(destValue))
                    returnProps.Add(prop.Name);
            }
            return returnProps;
        }
    }
}
